import json
from numpy.random import RandomState
from pathlib import Path
from typing import Self, Dict
import pandas as pd


class GenerativeModel:
    "Base class for all generative models"

    name: str = None

    def __init__(self, domain: Dict, random_state: RandomState = None):
        self.domain = domain
        self.random_state = random_state

    def set_domain(self, domain: Dict):
        self.domain = domain

    def set_random_state(self, random_state: RandomState):
        self.random_state = random_state

    def fit(self, df: pd.DataFrame):
        "Fit a generative model"
        return NotImplementedError("Method needs to be overwritten.")

    def check_fit(self, df: pd.DataFrame):
        "Fit a generative model"
        return NotImplementedError("Method needs to be overwritten.")

    def generate(self, n_records: int = None, condition_records: pd.DataFrame = None):
        "Generate a synthetic dataset of size n_records"
        return NotImplementedError("Method needs to be overwritten.")

    def store(self, path: Path):
        "Generate a synthetic dataset of size n_records"
        with (path / "model_type.json").open("w") as fw:
            json.dump(self.name, fw)

    @classmethod
    def load(cls, path: Path) -> Self:
        "Generate a synthetic dataset of size n_records"
        return NotImplementedError("Method needs to be overwritten.")



# Adapted from: https://github.com/ryan112358/private-pgm/blob/1da21c8b38149b05f1385b8e54116568b700b4fa/src/mbi/inference.py

import numpy as np

# from mbi import Domain
from dpmm.models.base.mbi import GraphicalModel, callbacks, CliqueVector

# from scipy.sparse.linalg import LinearOperator
from scipy.sparse.linalg import eigsh, lsmr, aslinearoperator

# from scipy import optimize
from scipy import sparse

# from functools import partial
from collections import defaultdict


class FactoredInference:
    def __init__(
        self,
        domain,
        backend="numpy",
        structural_zeros={},
        metric="L2",
        log=False,
        iters=1000,
        warm_start=False,
        elim_order=None,
        prng: np.random.RandomState = None,
    ):
        """
        Class for learning a GraphicalModel from  noisy measurements on a data distribution

        :param domain: The domain information (A Domain object)
        :param backend: numpy or torch backend
        :param structural_zeros: An encoding of the known (structural) zeros in the distribution.
            Specified as a dictionary where
                - each key is a subset of attributes of size r
                - each value is a list of r-tuples corresponding to impossible attribute settings
        :param metric: The optimization metric.  May be L1, L2 or a custom callable function
            - custom callable function must consume the marginals and produce the loss and gradient
            - see FactoredInference._marginal_loss for more information
        :param log: flag to log iterations of optimization
        :param iters: number of iterations to optimize for
        :param warm_start: initialize new model or reuse last model when calling infer multiple times
        :param elim_order: an elimination order for the JunctionTree algorithm
            - Elimination order will impact the efficiency by not correctness.
              By default, a greedy elimination order is used
        """
        self.domain = domain
        self.backend = backend
        self.metric = metric
        self.log = log
        self.iters = iters
        self.warm_start = warm_start
        self.history = []
        self.elim_order = elim_order
        self._prng = prng
        if backend == "torch":
            from mbi.torch_factor import Factor

            self.Factor = Factor
        else:
            from dpmm.models.base.mbi import Factor

            self.Factor = Factor

        self.structural_zeros = CliqueVector({})
        for cl in structural_zeros:
            dom = self.domain.project(cl)
            fact = structural_zeros[cl]
            self.structural_zeros[cl] = self.Factor.active(dom, fact)

    @property
    def prng(self):
        if self._prng is None:
            return np.random.RandomState()
        return self._prng

    def estimate(
        self, measurements, total=None, engine="MD", callback=None, options={}
    ):
        """
        Estimate a GraphicalModel from the given measurements

        :param measurements: a list of (Q, y, noise, proj) tuples, where
            Q is the measurement matrix (a numpy array or scipy sparse matrix or LinearOperator)
            y is the noisy answers to the measurement queries
            noise is the standard deviation of the noise added to y
            proj defines the marginal used for this measurement set (a subset of attributes)
        :param total: The total number of records (if known)
        :param engine: the optimization algorithm to use, options include:
            MD - Mirror Descent with armijo line search
            RDA - Regularized Dual Averaging
            IG - Interior Gradient
        :param callback: a function to be called after each iteration of optimization
        :param options: solver specific options passed as a dictionary
            { param_name : param_value }

        :return model: A GraphicalModel that best matches the measurements taken
        """
        measurements = self.fix_measurements(measurements)
        options["callback"] = callback
        if callback is None and self.log:
            options["callback"] = callbacks.Logger(self)
        if engine == "MD":
            self.mirror_descent(measurements, total, **options)
        elif engine == "RDA":
            self.dual_averaging(measurements, total, **options)
        elif engine == "IG":
            self.interior_gradient(measurements, total, **options)
        return self.model

    def fix_measurements(self, measurements):
        assert type(measurements) is list, (
            "measurements must be a list, given " + measurements
        )
        assert all(
            len(m) == 4 for m in measurements
        ), "each measurement must be a 4-tuple (Q, y, noise,proj)"
        ans = []
        for Q, y, noise, proj in measurements:
            assert (
                Q is None or Q.shape[0] == y.size
            ), "shapes of Q and y are not compatible"
            if type(proj) is list:
                proj = tuple(proj)
            if type(proj) is not tuple:
                proj = (proj,)
            if Q is None:
                Q = sparse.eye(self.domain.size(proj))
            assert np.isscalar(noise), "noise must be a real value, given " + str(noise)
            assert all(a in self.domain for a in proj), (
                str(proj) + " not contained in domain"
            )
            assert Q.shape[1] == self.domain.size(
                proj
            ), "shapes of Q and proj are not compatible"
            ans.append((Q, y, noise, proj))
        return ans

    def interior_gradient(
        self, measurements, total, lipschitz=None, c=1, sigma=1, callback=None
    ):
        """Use the interior gradient algorithm to estimate the GraphicalModel
            See https://epubs.siam.org/doi/pdf/10.1137/S1052623403427823 for more information

        :param measurements: a list of (Q, y, noise, proj) tuples, where
            Q is the measurement matrix (a numpy array or scipy sparse matrix or LinearOperator)
            y is the noisy answers to the measurement queries
            noise is the standard deviation of the noise added to y
            proj defines the marginal used for this measurement set (a subset of attributes)
        :param total: The total number of records (if known)
        :param lipschitz: the Lipchitz constant of grad L(mu)
            - automatically calculated for metric=L2
            - doesn't exist for metric=L1
            - must be supplied for custom callable metrics
        :param c, sigma: parameters of the algorithm
        :param callback: a function to be called after each iteration of optimization
        """
        assert self.metric != "L1", "dual_averaging cannot be used with metric=L1"
        assert (
            not callable(self.metric) or lipschitz is not None
        ), "lipschitz constant must be supplied"
        self._setup(measurements, total)
        # what are c and sigma?  For now using 1
        model = self.model
        domain, cliques, total = model.domain, model.cliques, model.total
        L = self._lipschitz(measurements) if lipschitz is None else lipschitz
        if self.log:
            print("Lipchitz constant:", L)

        theta = model.potentials
        x = y = z = model.belief_propagation(theta)
        c0 = c
        l = sigma / L
        for k in range(1, self.iters + 1):
            a = (np.sqrt((c * l) ** 2 + 4 * c * l) - l * c) / 2
            y = (1 - a) * x + a * z
            c *= 1 - a
            _, g = self._marginal_loss(y)
            theta = theta - a / c / total * g
            z = model.belief_propagation(theta)
            x = (1 - a) * x + a * z
            if callback is not None:
                callback(x)

        model.marginals = x
        model.potentials = model.mle(x)

    def dual_averaging(self, measurements, total=None, lipschitz=None, callback=None):
        """Use the regularized dual averaging algorithm to estimate the GraphicalModel
            See https://www.microsoft.com/en-us/research/wp-content/uploads/2016/02/xiao10JMLR.pdf

        :param measurements: a list of (Q, y, noise, proj) tuples, where
            Q is the measurement matrix (a numpy array or scipy sparse matrix or LinearOperator)
            y is the noisy answers to the measurement queries
            noise is the standard deviation of the noise added to y
            proj defines the marginal used for this measurement set (a subset of attributes)
        :param total: The total number of records (if known)
        :param lipschitz: the Lipchitz constant of grad L(mu)
            - automatically calculated for metric=L2
            - doesn't exist for metric=L1
            - must be supplied for custom callable metrics
        :param callback: a function to be called after each iteration of optimization
        """
        assert self.metric != "L1", "dual_averaging cannot be used with metric=L1"
        assert (
            not callable(self.metric) or lipschitz is not None
        ), "lipschitz constant must be supplied"
        self._setup(measurements, total)
        model = self.model
        domain, cliques, total = model.domain, model.cliques, model.total
        L = self._lipschitz(measurements) if lipschitz is None else lipschitz
        print("Lipchitz constant:", L)
        if L == 0:
            return

        theta = model.potentials
        gbar = CliqueVector(
            {cl: self.Factor.zeros(domain.project(cl)) for cl in cliques}
        )
        w = v = model.belief_propagation(theta)
        beta = 0

        for t in range(1, self.iters + 1):
            c = 2.0 / (t + 1)
            u = (1 - c) * w + c * v
            _, g = self._marginal_loss(u)  # not interested in loss of this query point
            gbar = (1 - c) * gbar + c * g
            theta = -t * (t + 1) / (4 * L + beta) / self.model.total * gbar
            v = model.belief_propagation(theta)
            w = (1 - c) * w + c * v

            if callback is not None:
                callback(w)

        model.marginals = w
        model.potentials = model.mle(w)

    def mirror_descent(self, measurements, total=None, stepsize=None, callback=None):
        """Use the mirror descent algorithm to estimate the GraphicalModel
            See https://web.iem.technion.ac.il/images/user-files/becka/papers/3.pdf

        :param measurements: a list of (Q, y, noise, proj) tuples, where
            Q is the measurement matrix (a numpy array or scipy sparse matrix or LinearOperator)
            y is the noisy answers to the measurement queries
            noise is the standard deviation of the noise added to y
            proj defines the marginal used for this measurement set (a subset of attributes)
        :param stepsize: The step size function for the optimization (None or scalar or function)
            if None, will perform line search at each iteration (requires smooth objective)
            if scalar, will use constant step size
            if function, will be called with the iteration number
        :param total: The total number of records (if known)
        :param callback: a function to be called after each iteration of optimization
        """
        assert not (
            self.metric == "L1" and stepsize is None
        ), "loss function not smooth, cannot use line search (specify stepsize)"

        self._setup(measurements, total)
        model = self.model
        cliques, theta = model.cliques, model.potentials
        mu = model.belief_propagation(theta)
        ans = self._marginal_loss(mu)
        if ans[0] == 0:
            model.potentials = theta
            model.marginals = mu
            return ans[0]

        nols = stepsize is not None
        if np.isscalar(stepsize):
            alpha = float(stepsize)
            stepsize = lambda t: alpha
        if stepsize is None:
            alpha = 1.0 / self.model.total**2
            stepsize = lambda t: 2.0 * alpha

        for t in range(1, self.iters + 1):
            if callback is not None:
                callback(mu)
            omega, nu = theta, mu
            curr_loss, dL = ans
            # print('Gradient Norm', np.sqrt(dL.dot(dL)))
            alpha = stepsize(t)
            for i in range(25):
                theta = omega - alpha * dL
                mu = model.belief_propagation(theta)
                ans = self._marginal_loss(mu)
                if nols or curr_loss - ans[0] >= 0.5 * alpha * dL.dot(nu - mu):
                    break
                alpha *= 0.5

        model.potentials = theta
        model.marginals = mu

        return ans[0]

    def _marginal_loss(self, marginals, metric=None):
        """Compute the loss and gradient for a given dictionary of marginals

        :param marginals: A dictionary with keys as projections and values as Factors
        :return loss: the loss value
        :return grad: A dictionary with gradient for each marginal
        """
        if metric is None:
            metric = self.metric

        if callable(metric):
            return metric(marginals)

        loss = 0.0
        gradient = {}

        for cl in marginals:
            mu = marginals[cl]
            gradient[cl] = self.Factor.zeros(mu.domain)
            for Q, y, noise, proj in self.groups[cl]:
                c = 1.0 / noise
                mu2 = mu.project(proj)
                x = mu2.datavector()
                diff = c * (Q @ x - y)
                if metric == "L1":
                    loss += abs(diff).sum()
                    sign = diff.sign() if hasattr(diff, "sign") else np.sign(diff)
                    grad = c * (Q.T @ sign)
                else:
                    loss += 0.5 * (diff @ diff)
                    grad = c * (Q.T @ diff)
                gradient[cl] += self.Factor(mu2.domain, grad)
        return float(loss), CliqueVector(gradient)

    def _setup(self, measurements, total):
        """Perform necessary setup for running estimation algorithms

        1. If total is None, find the minimum variance unbiased estimate for total and use that
        2. Construct the GraphicalModel
            * If there are structural_zeros in the distribution, initialize factors appropriately
        3. Pre-process measurements into groups so that _marginal_loss may be evaluated efficiently
        """
        if total is None:
            # find the minimum variance estimate of the total given the measurements
            variances = np.array([])
            estimates = np.array([])
            for Q, y, noise, proj in measurements:
                o = np.ones(Q.shape[1])
                v = lsmr(Q.T, o, atol=0, btol=0)[0]
                if np.allclose(Q.T.dot(v), o):
                    variances = np.append(variances, noise**2 * np.dot(v, v))
                    estimates = np.append(estimates, np.dot(v, y))
            if estimates.size == 0:
                total = 1
            else:
                variance = 1.0 / np.sum(1.0 / variances)
                estimate = variance * np.sum(estimates / variances)
                total = max(1, estimate)

        # if not self.warm_start or not hasattr(self, 'model'):
        # initialize the model and parameters
        cliques = [m[3] for m in measurements]
        if self.structural_zeros is not None:
            cliques += list(self.structural_zeros.keys())

        model = GraphicalModel(
            self.domain,
            cliques,
            total,
            elimination_order=self.elim_order,
            prng=self.prng,
        )

        model.potentials = CliqueVector.zeros(self.domain, model.cliques)
        model.potentials.combine(self.structural_zeros)
        if self.warm_start and hasattr(self, "model"):
            model.potentials.combine(self.model.potentials)

        self.model = model

        # group the measurements into model cliques
        cliques = self.model.cliques
        # self.groups = { cl : [] for cl in cliques }
        self.groups = defaultdict(list)
        for Q, y, noise, proj in measurements:
            if self.backend == "torch":
                import torch

                device = self.Factor.device
                y = torch.tensor(y, dtype=torch.float32, device=device)
                if isinstance(Q, np.ndarray):
                    Q = torch.tensor(Q, dtype=torch.float32, device=device)
                elif sparse.issparse(Q):
                    Q = Q.tocoo()
                    idx = torch.LongTensor([Q.row, Q.col])
                    vals = torch.FloatTensor(Q.data)
                    Q = torch.sparse.FloatTensor(idx, vals).to(device)

                # else Q is a Linear Operator, must be compatible with torch
            m = (Q, y, noise, proj)
            for cl in sorted(cliques, key=model.domain.size):
                # (Q, y, noise, proj) tuple
                if set(proj) <= set(cl):
                    self.groups[cl].append(m)
                    break

    def _lipschitz(self, measurements):
        """compute lipschitz constant for L2 loss

        Note: must be called after _setup
        """
        eigs = {cl: 0.0 for cl in self.model.cliques}
        for Q, _, noise, proj in measurements:
            for cl in self.model.cliques:
                if set(proj) <= set(cl):
                    n = self.domain.size(cl)
                    p = self.domain.size(proj)
                    Q = aslinearoperator(Q)
                    Q.dtype = np.dtype(Q.dtype)
                    eig = eigsh(Q.H * Q, 1)[0][0]
                    eigs[cl] += eig * n / p / noise**2
                    break
        return max(eigs.values())

    def infer(self, measurements, total=None, engine="MD", callback=None, options={}):
        import warnings

        message = "Function infer is deprecated.  Please use estimate instead."
        warnings.warn(message, DeprecationWarning)
        return self.estimate(measurements, total, engine, callback, options)



# Implementation of AIM: An Adaptive and Iterative Mechanism for DP Synthetic Data.
# Note that with the default settings, AIM can take many hours to run.  You can configure
# the runtime /utility tradeoff via the max_model_size flag.  We recommend setting it to 1.0
# for debugging, but keeping the default value of 80 for any official comparisons to this mechanism.
# Note that we assume in this file that the data has been appropriately preprocessed so that there are no large-cardinality categorical attributes.  If there are, we recommend using something like "compress_domain" from mst.py.  Since our paper evaluated already-preprocessed datastes, we did not implement that here for simplicity.

# Adapted from: https://github.com/ryan112358/private-pgm/blob/master/mechanisms/aim.py

import itertools
from multiprocessing import Pool
from pathlib import Path
from typing import Dict, Self

import joblib
import numpy as np
import pandas as pd
from numpy.random import RandomState
from scipy import sparse

from dpmm.models.base.graphical import GraphicalGenerativeModel
from dpmm.models.base.matrix import Identity
from dpmm.models.base.mbi import Dataset, Domain, FactoredInference, GraphicalModel
from dpmm.models.base.mechanisms import Mechanism
from dpmm.models.base.utils import gaussian_noise
from dpmm.utils import to_path
from scipy.special import softmax


def powerset(iterable: list) -> itertools.chain:
    """
    Generate all non-empty subsets of an iterable.

    :param iterable: The input list.
    :type iterable: list
    :return: A chain object containing all subsets.
    :rtype: itertools.chain

    **Example**::

        >>> list(powerset([1, 2, 3]))
        [(1,), (2,), (3,), (1, 2), (1, 3), (2, 3), (1, 2, 3)]
    """
    s = list(iterable)
    return itertools.chain.from_iterable(
        itertools.combinations(s, r) for r in range(1, len(s) + 1)
    )


def downward_closure(Ws: list) -> list:
    """
    Compute the downward closure of a set of projections.

    :param Ws: A list of projections.
    :type Ws: list
    :return: A sorted list of subsets in the downward closure.
    :rtype: list

    **Example**::

        >>> downward_closure([(1, 2), (2, 3)])
        [(1,), (2,), (3,), (1, 2), (2, 3)]
    """
    ans = set()
    for proj in Ws:
        ans.update(powerset(proj))
    return list(sorted(ans, key=len))


def hypothetical_model_size(domain: Domain, cliques: list) -> float:
    """
    Calculate the hypothetical size of a graphical model.

    :param domain: The domain of the model.
    :type domain: Domain
    :param cliques: A list of cliques.
    :type cliques: list
    :return: The size of the model in megabytes.
    :rtype: float

    **Example**::

        >>> hypothetical_model_size(domain, [(1, 2), (2, 3)])
        0.015625
    """
    model = GraphicalModel(domain, cliques)
    return model.size * 8 / 2**20


def compile_workload(workload: list) -> Dict[tuple, int]:
    """
    Compile a workload into a dictionary of cliques and their scores.

    :param workload: A list of workload cliques.
    :type workload: list
    :return: A dictionary mapping cliques to scores.
    :rtype: Dict[tuple, int]

    **Example**::

        >>> compile_workload([(1, 2), (2, 3)])
        {(1,): 1, (2,): 2, (3,): 1, (1, 2): 2, (2, 3): 2}
    """

    def score(cl):
        return sum(len(set(cl) & set(ax)) for ax in workload)

    return {cl: score(cl) for cl in downward_closure(workload)}


def filter_candidates(
    candidates: Dict[tuple, int], model: GraphicalModel, size_limit: float
) -> Dict[tuple, int]:
    """
    Filter candidates based on size constraints and free cliques.

    :param candidates: Candidate cliques with scores.
    :type candidates: Dict[tuple, int]
    :param model: The current graphical model.
    :type model: GraphicalModel
    :param size_limit: The maximum allowed model size.
    :type size_limit: float
    :return: Filtered candidates.
    :rtype: Dict[tuple, int]

    **Example**::

        >>> filter_candidates(candidates, model, 80)
        {(1, 2): 3, (2, 3): 2}
    """
    ans = {}
    free_cliques = downward_closure(model.cliques)
    for cl in candidates:
        if size_limit is None:
            cond1 = True
        else:
            cond1 = (
                hypothetical_model_size(model.domain, model.cliques + [cl])
                <= size_limit
            )
        cond2 = cl in free_cliques
        if cond1 or cond2:
            ans[cl] = candidates[cl]
    return ans


def measure_one_way(
    cl: tuple, data: Dataset, sigma: float, prng: RandomState, public: bool = False
) -> tuple:
    """
    Measure a single clique with optional noise.

    :param cl: The clique to measure.
    :type cl: tuple
    :param data: The dataset.
    :type data: Dataset
    :param sigma: The noise scale.
    :type sigma: float
    :param prng: The random state.
    :type prng: RandomState
    :param public: Whether the data is public. Defaults to False.
    :type public: bool, optional
    :return: Measurement matrix, noisy data vector, noise scale, and clique.
    :rtype: tuple

    **Example**::

        >>> measure_one_way((1, 2), data, 1.0, prng)
        (<Identity>, array([0.5, 0.3]), 1.0, (1, 2))
    """
    x = data.project(cl).datavector()
    if public:
        y = x
    else:
        y = x + gaussian_noise(sigma=sigma, size=x.size)
    I = Identity(y.size)
    return (I, y, sigma, cl)


def _measure(
    data: Dataset, proj: tuple, wgt: float, sigma: float, public: bool = False
) -> tuple:
    """
    Perform a measurement for a single projection.

    :param data: The dataset.
    :type data: Dataset
    :param proj: The projection.
    :type proj: tuple
    :param wgt: The weight.
    :type wgt: float
    :param sigma: The noise scale.
    :type sigma: float
    :param public: Whether the data is public. Defaults to False.
    :type public: bool, optional
    :return: Measurement matrix, noisy data vector, noise scale, and projection.
    :rtype: tuple

    **Example**::

        >>> _measure(data, (1, 2), 0.5, 1.0)
        (<sparse.eye>, array([0.5, 0.3]), 2.0, (1, 2))
    """
    x = data.project(proj).datavector()
    if public:
        y = x
    else:
        y = x + gaussian_noise(sigma=sigma, size=x.size)
    Q = sparse.eye(x.size)
    return (Q, y, sigma, proj)


def measure(
    data: Dataset,
    cliques: list,
    sigma: float,
    weights: np.ndarray = None,
    public: bool = False,
    n_jobs: int = -1,
) -> list:
    """
    Measure multiple cliques in parallel.

    :param data: The dataset.
    :type data: Dataset
    :param cliques: A list of cliques.
    :type cliques: list
    :param sigma: The noise scale.
    :type sigma: float
    :param weights: Weights for the cliques. Defaults to None.
    :type weights: np.ndarray, optional
    :param public: Whether the data is public. Defaults to False.
    :type public: bool, optional
    :param n_jobs: Number of parallel jobs. Defaults to -1.
    :type n_jobs: int, optional
    :return: A list of measurements.
    :rtype: list

    **Example**::

        >>> measure(data, [(1, 2), (2, 3)], 1.0)
        [(<sparse.eye>, array([0.5, 0.3]), 1.0, (1, 2)), ...]
    """
    if weights is None:
        weights = np.ones(len(cliques))
    weights = np.array(weights) / np.linalg.norm(weights)
    if n_jobs == 1:
        measurements = [_measure(*args) for args in zip(
                itertools.cycle([data]),
                cliques,
                weights,
                itertools.cycle([sigma]),
                itertools.cycle([public]),
            )]
    else:
        measurements = [
            meas
            for meas in Pool(n_jobs).starmap(
                _measure,
                zip(
                    itertools.cycle([data]),
                    cliques,
                    weights,
                    itertools.cycle([sigma]),
                    itertools.cycle([public]),
                ),
            )
        ]
    return measurements


class AIM(Mechanism):
    """
    Adaptive and Iterative Mechanism (AIM) for differential privacy.
    The model works by iteratively selecting the worst approximated clique
    and adding noise to it.
    It uses the exponential mechanism to select the worst approximated clique
    based on the quality scores of the candidates.

    Ref: https://arxiv.org/pdf/2201.12677


    :param epsilon: Privacy budget.
    :type epsilon: float
    :param delta: Privacy parameter.
    :type delta: float
    :param prng: Random state for reproducibility.
    :type prng: RandomState
    :param rounds: Number of rounds.
    :type rounds: int, optional
    :param max_model_size: Maximum model size in MB.
    :type max_model_size: float
    :param n_iters: Number of iterations for inference.
    :type n_iters: int
    :param degree: Maximum clique size.
    :type degree: int
    :param num_marginals: Number of marginals to sample.
    :type num_marginals: int, optional
    :param max_cells: Maximum number of cells in a clique.
    :type max_cells: int
    :param structural_zeros: Structural zeros in the data.
    :type structural_zeros: dict
    :param compress: Whether to compress the data.
    :type compress: bool
    :param domain: The domain of the data.
    :type domain: Domain
    :param n_jobs: Number of parallel jobs.
    :type n_jobs: int
    """

    def __init__(
        self,
        epsilon=1,
        delta=1e-5,
        prng: RandomState = None,
        rounds=None,
        max_model_size=80,
        n_iters: int = 1000,
        degree=2,
        num_marginals=None,
        max_cells=10000,
        structural_zeros={},
        compress=False,
        domain=None,
        n_jobs=-1,
    ):
        super().__init__(
            epsilon=epsilon,
            delta=delta,
            prng=prng,
            max_model_size=max_model_size,
            compress=compress,
            domain=domain,
            structural_zeros=structural_zeros,
            n_jobs=n_jobs,
        )

        self.rounds = rounds
        self.degree = degree
        self.n_iters = n_iters
        self.num_marginals = num_marginals
        self.max_cells = max_cells

    def _worst_approximated(self, cl, wgt, x, model, sigma):
        bias = np.sqrt(2 / np.pi) * sigma * model.domain.size(cl)
        xest = model.project(cl).datavector()
        error = wgt * (np.linalg.norm(x - xest, 1) - bias)
        sensitivity = abs(wgt)
        return cl, error, sensitivity

    def exponential_mechanism(
        self, qualities, epsilon, sensitivity=1.0, base_measure=None
    ):
        """
        Sample a candidate using the provided scores using the exponential mechanism.

        :param qualities: Quality scores for candidates.
        :type qualities: Union[dict, np.ndarray]
        :param epsilon: Privacy budget.
        :type epsilon: float
        :param sensitivity: Sensitivity of the scores. Defaults to 1.0.
        :type sensitivity: float, optional
        :param base_measure: Base measure for candidates. Defaults to None.
        :type base_measure: np.ndarray, optional
        :return: The selected candidate.
        :rtype: Any

        **Example**::

            >>> exponential_mechanism(qualities, 1.0)
            (1, 2)
        """
        if isinstance(qualities, dict):
            keys = list(qualities.keys())
            qualities = np.array([qualities[key] for key in keys])
            if base_measure is not None:
                base_measure = np.log([base_measure[key] for key in keys])
        else:
            qualities = np.array(qualities)
            keys = np.arange(qualities.size)

        q = qualities - qualities.max()
        if base_measure is None:
            p = softmax(0.5 * epsilon / sensitivity * q)
        else:
            p = softmax(0.5 * epsilon / sensitivity * q + base_measure)

        return keys[self.prng.choice(p.size, p=p)]

    def worst_approximated(self, candidates, answers, model, eps, sigma, public=False):
        """
        Find the worst approximated clique using the exponential mechanism.

        :param candidates: Candidate cliques with scores.
        :type candidates: dict
        :param answers: True answers for the cliques.
        :type answers: dict
        :param model: The current graphical model.
        :type model: GraphicalModel
        :param eps: Privacy budget.
        :type eps: float
        :param sigma: Noise scale.
        :type sigma: float
        :param public: Whether the data is public. Defaults to False.
        :type public: bool, optional
        :return: The worst approximated clique.
        :rtype: tuple

        **Example**::

            >>> worst_approximated(candidates, answers, model, 1.0, 0.5)
            (1, 2)
        """
        errors = {}
        sensitivity = {}

        if self.n_jobs == 1:
            for cl in candidates:
                _, errors[cl], sensitivity[cl] = self._worst_approximated(cl, candidates[cl], answers[cl], model, sigma)
        else:
            for cl, err, sens in Pool(self.n_jobs).starmap(
                self._worst_approximated,
                [(cl, candidates[cl], answers[cl], model, sigma) for cl in candidates],
            ):
                errors[cl] = err
                sensitivity[cl] = sens

        max_sensitivity = max(
            sensitivity.values()
        )  # if all weights are 0, could be a problem

        if public:
            _cliques = list(errors.keys())
            worst = np.argmax(errors.values())
            return _cliques[worst]
        else:
            return self.exponential_mechanism(errors, eps, max_sensitivity)

    def _fit(self, data: Dataset, public=False, W=None):
        """
        Fit the AIM mechanism to the data.

        :param data: The dataset.
        :type data: Dataset
        :param public: Whether the data is public. Defaults to False.
        :type public: bool, optional
        :param W: Workload cliques. Defaults to None.
        :type W: list, optional
        :return: The dataset and measurements.
        :rtype: tuple

        **Example**::

            >>> _fit(data, public=True)
            (data, measurements)
        """
        if W is None:
            W = list(itertools.combinations(data.domain, self.degree))
            W = [cl for cl in W if data.domain.size(cl) <= self.max_cells]
            W = [(cl, 1.0) for cl in W]

        if self.num_marginals is not None:
            W = [
                W[i]
                for i in self.prng.choice(len(W), self.num_marginals, replace=False)
            ]

        rounds = self.rounds or 16 * len(data.domain)
        workload = [cl for cl, _ in W]
        candidates = compile_workload(workload)

        oneway = [cl for cl in candidates if len(cl) == 1]

        sigma = np.sqrt(rounds / (2 * 0.9 * self.rho))
        epsilon = np.sqrt(8 * 0.1 * self.rho / rounds)

        measurements = []

        # cliques & sigmas
        self.cliques = list(oneway)

        rho_used = len(oneway) * 0.5 / sigma**2
        if self.n_jobs > 1:
            with Pool(self.n_jobs) as p:
                oneway_measurements = p.starmap(
                    measure_one_way,
                    [(cl, data, sigma, self.prng, public) for cl in oneway],
                )
            measurements.extend(oneway_measurements)
        else:
            measurements = [
                measure_one_way(cl, data.project(cl), sigma, self.prng, public)
                for cl in oneway
            ]

        if self.compress:
            measurements = self.compressor.fit(measurements)
            data = self.compressor.transform(data)

        self.engine = FactoredInference(
            data.domain,
            iters=self.n_iters,
            warm_start=True,
            prng=self.prng,
        )

        answers = {cl: data.project(cl).datavector() for cl in candidates}

        model = self.engine.estimate(measurements)

        t = 0
        terminate = False
        while not terminate:
            t += 1
            if public and t >= rounds:
                break
            elif self.rho - rho_used < 2 * (0.5 / sigma**2 + 1.0 / 8 * epsilon**2):
                # Just use up whatever remaining budget there is for one last round
                remaining = self.rho - rho_used
                if remaining < 0:
                    break
                sigma = np.sqrt(1 / (2 * 0.9 * remaining))
                epsilon = np.sqrt(8 * 0.1 * remaining)
                terminate = True

            rho_used += 1.0 / 8 * epsilon**2 + 0.5 / sigma**2
            if self.max_model_size is not None:
                size_limit = self.max_model_size * rho_used / self.rho
            else:
                size_limit = None

            small_candidates = filter_candidates(candidates, model, size_limit)
            cl = self.worst_approximated(
                small_candidates, answers, model, epsilon, sigma, public
            )

            n = data.domain.size(cl)
            Q = Identity(n)
            x = data.project(cl).datavector()
            # TODO: figure out to make determinstic
            if public:
                y = x
            else:
                y = x + gaussian_noise(sigma=sigma, size=n)

            self.cliques.append(cl)
            measurements.append((Q, y, sigma, cl))
            z = model.project(cl).datavector()

            model = self.engine.estimate(measurements)
            w = model.project(cl).datavector()
            if np.linalg.norm(w - z, 1) <= sigma * np.sqrt(2 / np.pi) * n:
                sigma /= 2
                epsilon *= 2

        return data, measurements

    def store(self, path: Path):
        """
        Store the AIM mechanism state to a file.

        :param path: The path to store the state.
        :type path: Path

        **Example**::

            >>> store(Path("/tmp/state"))
        """
        joblib.dump(
            {
                "epsilon": self.epsilon,
                "delta": self.delta,
                "rounds": self.rounds,
                "n_iters": self.n_iters,
                "max_model_size": self.max_model_size,
                "degree": self.degree,
                "num_marginals": self.num_marginals,
                "max_cells": self.max_cells,
                "structural_zeros": self.structural_zeros,
                "_domain": self._domain,
                "compress": self._compress,
                "compressor": self.compressor,
                "cliques": self.cliques,
                "fit_state": self.fit_state,
            },
            path / "state.joblib",
        )

        if self.model is not None:
            self.model.save(path / "estimator.pickle")

    @classmethod
    def load(cls, path: Path) -> Self:
        """
        Load the AIM mechanism state from a file.

        :param path: The path to load the state.
        :type path: Path
        :return: The loaded AIM mechanism.
        :rtype: AIM

        **Example**::

            >>> AIM.load(Path("/tmp/state"))
        """
        state = joblib.load(path / "state.joblib")
        obj = cls(
            epsilon=state["epsilon"],
            delta=state["delta"],
            rounds=state["rounds"],
            n_iters=state["n_iters"],
            max_model_size=state["max_model_size"],
            degree=state["degree"],
            num_marginals=state["num_marginals"],
            max_cells=state["max_cells"],
            structural_zeros=state["structural_zeros"],
            domain=state["_domain"],
            compress=state["compress"],
        )
        obj.fit_state = state["fit_state"]
        obj.cliques = state["cliques"]
        obj.compressor = state["compressor"]
        model_path = path / "estimator.pickle"
        if model_path.exists():
            obj.model = GraphicalModel.load(model_path)
        return obj


class AIMGM(GraphicalGenerativeModel):
    """
    Adaptive and Iterative Mechanism (AIM) for differential privacy.
    The model works by iteratively selecting the worst approximated clique
    and adding noise to it.
    It uses the exponential mechanism to select the worst approximated clique
    based on the quality scores of the candidates.

    Ref: https://arxiv.org/pdf/2201.12677

    :param epsilon: Privacy budget.
    :type epsilon: float
    :param delta: Privacy parameter.
    :type delta: float
    :param rounds: Number of rounds.
    :type rounds: int, optional
    :param compress: Whether to compress the data.
    :type compress: bool
    :param n_iters: Number of iterations for inference.
    :type n_iters: int
    :param max_model_size: Maximum model size in MB.
    :type max_model_size: float
    :param degree: Maximum clique size.
    :type degree: int
    :param num_marginals: Number of marginals to sample.
    :type num_marginals: int, optional
    :param max_cells: Maximum number of cells in a clique.
    :type max_cells: int
    :param structural_zeros: Structural zeros in the data.
    :type structural_zeros: dict
    :param domain: The domain of the data.
    :type domain: Domain
    :param random_state: Random state for reproducibility.
    :type random_state: RandomState
    """

    name = "aim"

    def __init__(
        self,
        epsilon=1,
        delta=1e-5,
        rounds=None,
        compress=True,
        n_iters: int = 1000,
        max_model_size=80,
        degree=2,
        num_marginals=None,
        max_cells=10000,
        structural_zeros={},
        domain=None,
        random_state: RandomState = None,
        n_jobs: int = -1,
    ):

        super().__init__(domain=domain, random_state=random_state)
        self.epsilon = epsilon
        self.delta = delta

        self.generator = AIM(
            epsilon=epsilon,
            delta=delta,
            prng=random_state,
            n_iters=n_iters,
            rounds=rounds,
            compress=compress,
            max_model_size=max_model_size,
            degree=degree,
            num_marginals=num_marginals,
            max_cells=max_cells,
            structural_zeros=structural_zeros,
            domain=domain,
            n_jobs=n_jobs,
        )

    def set_random_state(self, random_state: RandomState):
        """
        Set the random state for reproducibility.

        :param random_state: The random state.
        :type random_state: RandomState

        **Example**::

            >>> set_random_state(RandomState(42))
        """
        super().set_random_state(random_state)
        self.generator.set_random_state(random_state)

    def set_domain(self, domain: Dict):
        """
        Set the domain of the data.

        :param domain: The domain.
        :type domain: Dict

        **Example**::

            >>> set_domain({"age": [0, 100], "income": [0, 100000]})
        """
        super().set_domain(domain=domain)
        self.generator.set_domain(domain=domain)

    @to_path
    def store(self, path: Path):
        """
        Store the AIMGM model state to a file.

        :param path: The path to store the state.
        :type path: Path

        **Example**::

            >>> store(Path("/tmp/state"))
        """
        super().store(path)
        self.generator.store(path)

    @classmethod
    @to_path
    def load(cls, path: Path) -> Self:
        """
        Load the AIMGM model state from a file.

        :param path: The path to load the state.
        :type path: Path
        :return: The loaded AIMGM model.
        :rtype: AIMGM

        **Example**::

            >>> AIMGM.load(Path("/tmp/state"))
        """
        generator = AIM.load(path)
        obj = cls(
            epsilon=generator.epsilon,
            delta=generator.delta,
            rounds=generator.rounds,
            max_model_size=generator.max_model_size,
            degree=generator.degree,
            num_marginals=generator.num_marginals,
            max_cells=generator.max_cells,
            structural_zeros=generator.structural_zeros,
            domain=generator._domain,
            compress=generator.compress,
        )
        del obj.generator
        obj.generator = generator

        return obj



from pathlib import Path
from typing import Dict, Optional, Self, Union

import pandas as pd
from numpy.random import RandomState

from dpmm.models import load_model
from dpmm.models.base.base import GenerativeModel
from dpmm.processing.table_binner import TableBinner
from dpmm.utils import to_path


class GenerativePipeline:
    def __init__(self, gen: GenerativeModel, proc: Optional[TableBinner] = None):
        """
        Initialize A GenerativePipeline object. 
        This object is used to chain together a generative model and a preprocessing step. 
        Ensuring that generative model can ingest any data type [e.g. categorical, numerical, etc.]
        This will also ensure that the domain and structural zeros are set correctly for the generative model.

        :param gen: The generative model.
        :param proc: Optional table binner for preprocessing.

        .. code-block:: python
            from dpmm.models.mst import MSTGM
            from dpmm.processing.table_binner import TableBinner
            model = MSTGM(epsilon = 1.0, delta = 1e-5)
            binner = TableBinner()
            pipeline = GenerativePipeline(gen=model, proc=binner)
        """
        self.gen = gen
        self.proc = proc
        self.random_state = None

    def set_random_state(self, rnd: Union[int, RandomState] = None) -> None:
        """
        Set the random state for reproducibility.

        :param rnd: An integer seed or a RandomState instance.

        .. code-block:: python

            pipeline.set_random_state(42)
        """
        if not isinstance(rnd, RandomState):
            rnd = RandomState(rnd)

        self.random_state = rnd
        if self.proc is not None:
            self.proc.set_random_state(rnd)
        self.gen.set_random_state(rnd)

    def fit(
        self,
        df: pd.DataFrame,
        domain: Optional[Dict] = None,
        structural_zeros: Optional[Dict] = None,
        random_state: Union[int, RandomState] = None,
        public: bool = False,
        marginals_only: bool = False
    ) -> None:
        """
        Fit the pipeline to the data.

        :param df: The input DataFrame.
        :param domain: Optional domain information for the data.
        :param structural_zeros: Optional structural zeros for specific columns.
        :param random_state: An integer seed or a RandomState instance.
        :param public: Whether the data is public (no privacy concerns).

        .. code-block:: python

            pipeline.fit(df, domain={"col1": {"lower": 0, "upper": 10}})
        """
        self.set_random_state(random_state)
        # Processing
        if self.proc is not None:
            if not (self.proc.is_fit):
                self.proc.set_domain(domain)
                t_df = self.proc.fit_transform(df, public=public)
            else:
                t_df = self.proc.transform(df)

            if structural_zeros is not None:
                self.proc.set_structural_zeros(structural_zeros)
            zeros = self.proc.zeros
            t_domain = self.proc.bin_domain
        else:
            t_domain = domain
            t_df = df
            zeros = structural_zeros

        # Generation
        self.gen.set_domain(domain=t_domain)
        self.gen.check_fit(t_df)
        if zeros is not None:
            self.gen.set_structural_zeros(zeros)
        self.gen.fit(t_df, public=public, marginals_only=marginals_only)

    def generate(
        self,
        n_records: Optional[int] = None,
        condition_records: Optional[pd.DataFrame] = None,
        random_state: Union[int, RandomState] = None,
    ) -> pd.DataFrame:
        """
        Generate synthetic data.

        :param n_records: Number of records to generate.
        :param condition_records: Optional DataFrame to condition the generation.
        :param random_state: An integer seed or a RandomState instance.
        :return: A DataFrame of synthetic data.

        .. code-block:: python

            synthetic_data = pipeline.generate(n_records=100)
        """
        msg = "InvalidInput: Either 'n_records' or 'condition_records' must be set, both provided as None."
        assert (condition_records is not None) or (n_records is not None), msg

        if condition_records is not None:
            n_records = condition_records.shape[0]
            t_condition = self.proc.transform(condition_records)
        else:
            t_condition = None
        self.set_random_state(random_state)
        synth_df = self.gen.generate(n_records=n_records, condition_records=t_condition)

        if self.proc is not None:
            synth_df = self.proc.inverse_transform(df=synth_df)

        if condition_records is not None:
            synth_df = pd.concat(
                [condition_records, synth_df.drop(columns=condition_records.columns)],
                axis=1,
            )
        return synth_df

    @to_path
    def store(self, path: Path) -> None:
        """
        Store the pipeline to a file.

        :param path: The file path to store the pipeline.

        .. code-block:: python

            pipeline.store("pipeline_path")
        """
        # Processing
        if self.proc is not None:
            proc_path = path / "processing.joblib"
            self.proc.store(proc_path)

        # Generation
        gen_path = path / "generative_model"
        gen_path.mkdir(exist_ok=True, parents=True)
        self.gen.store(gen_path)

    @classmethod
    @to_path
    def load(cls, path: Path) -> Self:
        """
        Load the pipeline from a file.

        :param path: The file path to load the pipeline from.
        :return: The loaded GenerativePipeline instance.

        .. code-block:: python

            pipeline = GenerativePipeline.load("pipeline_path")
        """
        proc_path = path / "processing.joblib"
        if proc_path.exists():
            proc = TableBinner.load(proc_path)
        else:
            proc = None

        gen = load_model(path / "generative_model")
        return GenerativePipeline(gen=gen, proc=proc)


class MMPipeline(GenerativePipeline):
    model: GenerativeModel = None

    def __init__(
        self,
        # Model Params
        epsilon: float = 1,
        delta: float = 1e-5,
        compress: bool = True,
        max_model_size: Optional[int] = None,
        n_jobs: int = -1,
        gen_kwargs: Optional[Dict] = None,
        # Processing Params
        binner_type: Optional[str] = None,
        proc_epsilon: float = 0.1,
        n_bins: Union[str, int] = "auto",
        binner_kwargs: Optional[Dict] = None,
        disable_processing: bool = False,
    ):
        """
        Initialize the MMPipeline.

        :param epsilon: Privacy budget for the generative model.
        :param delta: Delta parameter for differential privacy.
        :param gen_kwargs: Additional arguments for the generative model.
        :param binner_type: Type of binner to use (e.g., "priv-tree").
        :param proc_epsilon: Privacy budget for the processing step.
        :param n_bins: Number of bins or 'auto' for automatic binning.
        :param binner_kwargs: Additional arguments for the binner.
        :param disable_processing: Whether to disable the processing step.

        .. code-block:: python

            pipeline = MMPipeline(
                epsilon=1.0,
                delta=1e-5,
                binner_type="uniform",
                proc_epsilon=0.1,
                n_bins=10,
            )
        """
        if gen_kwargs is None:
            gen_kwargs = {}
        if binner_kwargs is None:
            binner_kwargs = {}

        if binner_type is None:
            if proc_epsilon is not None:
                binner_type = "priv-tree"
            else:
                binner_type = "uniform"

        gen = self.model(
            epsilon=epsilon, 
            delta=delta,
            compress=compress,
            max_model_size=max_model_size,
            n_jobs=n_jobs,
            **gen_kwargs)

        if disable_processing:
            proc = None
        else:
            proc = TableBinner(
                binner_type=binner_type,
                binner_settings=dict(
                    epsilon=proc_epsilon, n_bins=n_bins, **binner_kwargs
                ),
            )

        super().__init__(gen=gen, proc=proc)

    @property
    def model_size(self):
        return self.gen.model_size



"""
    Code from :
    - https://github.com/opendp/smartnoise-sdk/blob/10b152700749aa7dbe85f73d6382e6899ea256f6/synth/snsynth/transform/mechanism.py
    - https://desfontain.es/thesis/Usability.html

"""

import math

import numpy as np
import pandas as pd
from opendp.domains import atom_domain
from opendp.measurements import make_laplace
from opendp.metrics import absolute_distance
from opendp.mod import enable_features
from typing import List, Optional, Tuple, Union


STD_DT = pd.to_datetime("01.01.1990").to_numpy()
STD_TD = pd.to_timedelta("1 hour").to_numpy()


def approx_bounds(vals: Union[List[float], np.ndarray], epsilon: float) -> Tuple[Optional[float], Optional[float]]:
    """
    Estimate the minimum and maximum values of a list of values.
    from: https://desfontain.es/thesis/Usability.html#usability-u-ding-

    :param vals: A list of values. Must be numeric.
    :param epsilon: The privacy budget to spend estimating the bounds.
    :return: A tuple of the estimated minimum and maximum values.

    .. code-block:: python

        vals = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        lower, upper = approx_bounds(vals, 0.1)
    """

    def get_bins(n_bins):
        return (
            list(-1 * 2 ** np.arange(n_bins)[::-1]) + [0] + list(2 ** np.arange(n_bins))
        )

    n_bins = 32
    vals = np.array(vals)

    dkind = vals.dtype.kind
    if dkind == "M":
        # Standardise to float
        vals = (vals - STD_DT) / STD_TD
    elif dkind == "m":
        # Standardise to float
        vals = vals / STD_TD

    bins = get_bins(n_bins)
    hist, _ = np.histogram(vals[(vals >= bins[0]) & (vals < bins[-1])], bins=bins)

    def edges(idx):
        return bins[idx], bins[idx + 1]

    # add noise
    enable_features("floating-point", "contrib")

    discovered_scale = 1.0 / epsilon
    input_space = atom_domain(T=float), absolute_distance(T=float)
    meas = make_laplace(*input_space, scale=discovered_scale)
    hist = [meas(float(v)) for v in hist]
    n_bins = len(hist)

    failure_prob = 10e-9
    highest_failure_prob = 1 / (n_bins * 2)

    exceeds = []
    while len(exceeds) < 1 and failure_prob <= highest_failure_prob:
        p = 1 - failure_prob
        K = -np.log(2 - 2 * p ** (1 / (n_bins - 1))) / epsilon
        exceeds = [idx for idx, v in enumerate(hist) if v > K]
        failure_prob *= 10

    if len(exceeds) == 0:
        return (None, None)

    lower, upper = min(exceeds), max(exceeds)
    ll, _ = edges(lower)
    _, uu = edges(upper)

    if dkind == "M":
        ll, uu = pd.Timestamp(STD_DT + ll * STD_TD), pd.Timestamp(STD_DT + uu * STD_TD)
    elif dkind == "m":
        ll, uu = pd.Timedelta(ll * STD_TD), pd.Timedelta(uu * STD_TD)
    elif dkind == "f":
        ll, uu = float(ll), float(uu)
    return (ll, uu)


def optimal_n_bins(series: pd.Series, upper: Optional[int] = None, epsilon: Optional[float] = None) -> int:
    """Calculate the optimal number of bins for a histogram.

    :param series: A pandas Series containing the data.
    :param upper: An optional upper limit for the number of bins.
    :param epsilon: An optional privacy budget parameter.
    :return: The optimal number of bins as an integer.

    .. code-block:: python

        series = pd.Series([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        n_bins = optimal_n_bins(series, upper=20, epsilon=0.1)
    """
    n_series = len(series)
    n_bins = 2 * math.pow(n_series, 1 / 3)

    if pd.notnull(epsilon):
        n_bins *= 1 / (1 + np.exp(-epsilon))  # ].5 - 1[
        n_bins = max(n_bins, 5)

    if upper is not None:
        n_bins = min(n_bins, upper)
    return int(np.ceil(n_bins))



from dpmm.pipelines.aim import AIMPipeline
from dpmm.pipelines.mst import MSTPipeline
from dpmm.pipelines.priv_bayes import PrivBayesPipeline

PIPELINES = [AIMPipeline, MSTPipeline, PrivBayesPipeline]
PIPELINE_DICT = {PIPE.model.name: PIPE for PIPE in PIPELINES}



import pandas as pd
import numpy as np
from scipy import sparse
from dpmm.models.base.mbi import Dataset, Domain


class DomainCompressor:
    def __init__(self, prng=None, structural_zeros=None, zeros_only=False):
        self.supports = {}
        self.zeros = {}
        if prng is None:
            prng = np.random.RandomState()
        self.prng = prng
        # structural zeros
        if structural_zeros is None:
            structural_zeros = {}
        self.structural_zeros = structural_zeros
        self.zeros_only = zeros_only

    def fit(self, measurements, flatten=False):
        self.supports = {}
        return [self.fit_measure(measure, flatten=flatten) for measure in measurements]

    def fit_measure(self, measure, flatten=False):
        Q, y, sigma, proj = measure

        I2 = np.ones(y.size)

        for axis, col in enumerate(proj):
            y_col = np.sum(y, axis=tuple([ax for ax in range(len(proj)) if ax != axis]))

            if col not in self.supports:
                # Structural Zeros
                self.zeros[col] = np.zeros(y_col.shape[0], dtype=bool)
                if col in self.structural_zeros:
                    struct = np.array(self.structural_zeros[col]).astype(int)
                    self.zeros[col][struct] = True

                self.supports[col] = ~self.zeros[col]

                if not self.zeros_only:
                    if pd.isnull(sigma):
                        sup = np.ones(y_col.shape[0], dtype=bool)
                    else:
                        sup = y_col >= (3 * sigma)

                        if sup.sum() == 0:
                            # TODO: Add warning compression for target column failed
                            sup = np.ones(y_col.shape[0], dtype=bool)

                    self.supports[col] &= sup

            sup = self.supports[col]

            if self.supports[col].sum() == y.shape[axis]:
                continue
            # need to re-express measurement over the new domain
            else:
                # Select Counts to Preserve
                preserved_idx = np.arange(y_col.shape[0])[sup]
                y_new = np.take(y, indices=preserved_idx, axis=axis)

                # Concatenating Count
                if not self.zeros_only:
                    normaliser = np.sqrt(y_col.shape[0] - sup.sum() + 1)
                    agg_idx = np.arange(y_col.shape[0])[~sup]
                    y_agg = (
                        np.expand_dims(
                            np.take(y, indices=agg_idx, axis=axis).sum(axis=axis),
                            axis=axis,
                        )
                        / normaliser
                    )
                    y = np.concatenate((y_new, y_agg), axis=axis)
                else:
                    y = y_new

                # Correct Weight
                I2 = np.ones(y.size)
                if len(proj) == 1 and not self.zeros_only:
                    I2[-1] = 1.0 / normaliser

        Q = sparse.diags(I2)

        if flatten:
            y = y.flatten()

        return (Q, y, sigma, proj)

    def transform_col(self, df, col):
        as_df = isinstance(df, pd.DataFrame)
        if not as_df:
            domain = df.domain.config
            df = df.df

        support = self.supports[col]
        size = support.sum()
        newdom = int(size)

        mapped = df[col].copy()

        if col in self.zeros:
            # TODO : Add warning zeros only
            is_zero = self.zeros[col][mapped]
            if is_zero.any():
                mapped.loc[is_zero] = (
                    mapped.loc[~is_zero]
                    .sample(n=is_zero.sum(), replace=True)
                    .to_numpy()
                )

        mapping = {}
        idx = 0
        for i in range(support.size):
            mapping[i] = size
            if support[i]:
                mapping[i] = idx
                idx += 1
        assert idx == size

        mapped = mapped.astype(int).map(mapping)

        if (size < support.size) and (not self.zeros_only):
            newdom += 1

        if not as_df:
            domain[col] = newdom
            df[col] = mapped
            return Dataset(df=df, domain=Domain.fromdict(domain))

        return mapped, newdom

    def transform(self, data):
        as_df = isinstance(data, pd.DataFrame)
        if as_df:
            df = data.copy()
        else:
            df = data.df.copy()

        newdom = {}
        for col in df.columns:
            df[col], newdom[col] = self.transform_col(df, col)
        newdom = Domain.fromdict(newdom)
        if as_df:
            data = df
        else:
            data = Dataset(df, newdom)
        return data

    def reverse(self, data):
        df = data.df.copy()
        newdom = {}
        for col in data.domain:
            support = self.supports[col]
            if self.zeros_only:
                support_map = dict(
                    zip(np.arange(support.sum()), np.arange(support.size)[support])
                )
                df[col] = df[col].map(support_map)
                newdom[col] = support.size
            else:
                mx = support.sum()
                newdom[col] = int(support.size)
                agg = ~support
                if col in self.zeros:
                    agg &= ~self.zeros[col]
                idx, extra = np.where(support)[0], np.where(agg)[0]
                mask = df[col] == mx
                if extra.sum() == 0:
                    extra = idx

                df.loc[mask, col] = self.prng.choice(extra, mask.sum())
                df.loc[~mask, col] = idx[df.loc[~mask, col]]
        newdom = Domain.fromdict(newdom)
        return Dataset(df, newdom)



# Adapted from: https://github.com/ryan112358/private-pgm/blob/master/mechanisms/mechanism.py

import numpy as np
import pandas as pd

import itertools
from numpy.random import RandomState
from scipy import sparse
from typing import Dict

from multiprocessing import Pool, cpu_count
from dpmm.models.base.mbi import Dataset, Domain, FactoredInference
from dpmm.models.base.compression import DomainCompressor
from dpmm.models.base.utils import gaussian_noise
from dpmm.models.base.mechanisms import cdp_rho


class Mechanism:
    def __init__(
        self,
        epsilon,
        delta,
        prng: RandomState = None,
        max_model_size: int = None,
        compress=False,
        domain=None,
        structural_zeros: Dict = None,
        n_jobs: int = -1,
    ):
        """
        Base class for a mechanism.
        :param epsilon: privacy parameter
        :param delta: privacy parameter
        :param prng: pseudo random number generator
        """

        # adjust for non-DP
        if not epsilon:
            epsilon = 0
            delta = 0.99999  # values >= 1 break the code

        self.epsilon = epsilon
        self.delta = delta
        self.rho = 0 if delta == 0 else cdp_rho(epsilon, delta)
        self._prng = prng

        self._compress = compress

        if structural_zeros is None:
            structural_zeros = {}

        if self._compress or len(structural_zeros) > 0:
            self.compressor = DomainCompressor(
                prng=self.prng,
                structural_zeros=structural_zeros,
                zeros_only=not (self._compress),
            )
        else:
            self.compressor = None

        self.set_structural_zeros(structural_zeros)

        self._domain = domain
        self.max_model_size = max_model_size
        self.model_size = None
        self.model = None
        self.fit_state = None
        self.cliques = None
        if n_jobs == -1:
            n_jobs = cpu_count() - 1
        self.n_jobs = n_jobs
        self._measures = None

    @property
    def compress(self):
        return self.compressor is not None

    @property
    def prng(self):
        if self._prng is None:
            return np.random.RandomState()
        return self._prng

    def set_random_state(self, random_state: RandomState):
        self._prng = random_state
        if self.compress:
            self.compressor.prng = random_state

        if self.model is not None:
            self.model.set_random_state(random_state)

    def set_structural_zeros(self, structural_zeros: Dict):
        self.structural_zeros = structural_zeros
        if len(structural_zeros) > 0:
            if self.compressor is None:
                self.compressor = DomainCompressor(
                    prng=self.prng,
                    structural_zeros=structural_zeros,
                    zeros_only=not (self.compress),
                )
            else:
                self.compressor.structural_zeros = structural_zeros
        else:
            if self._compress is False:
                self.compressor = None

    def set_domain(self, domain: Dict):
        self._domain = domain

    def fit(self, df, public=False, marginals_only=False, *args, **kwargs):
        # prepare data
        if self._domain is None:
            _domain = (df.astype(int).max(axis=0) + 1).to_dict()
            if not public:
                # TODO: Add warning
                pass
            else:
                self._domain = _domain
        else:
            _domain = self._domain

        domain = Domain(
            list(df.columns), np.array([_domain[col] for col in df.columns])
        )

        data = Dataset(df, domain)

        # select cliques
        if (self.fit_state is None) or public:
            data, measures = self._fit(data=data, public=public, *args, **kwargs)
        else:
            if self.fit_state == "pretrained":
                self.rho = cdp_rho(self.epsilon, self.delta)
                self.sigma = np.sqrt(1 / (2 * self.rho))

            if self.compress:
                data = self.compressor.transform(data)

            measures = self.measure(data, public=public, flatten=True)

        # measures
        self._measures = measures
        self.model_size = sum([y.nbytes for (_, y, _, _) in measures]) / (8 * 1024**2)
        self.fit_state = "pretrained"
        if not public and not marginals_only:
            engine = FactoredInference(domain=data.domain, iters=self.n_iters, prng=self.prng)
            self.model = engine.estimate(measures)

            self.fit_state = "trained"

    @property
    def measures(self):
        return self._measures

    def _measure(self, data, proj, wgt, flatten=None, public=False):
        if flatten is None:
            flatten = self.compress
        x = data.project(list(proj)).datavector(flatten=flatten)
        # TODO: figure out to make determinstic
        if public:
            y = x
        else:
            y = x + gaussian_noise(sigma=self.sigma / wgt, size=x.shape)
        Q = sparse.eye(x.size)
        return (Q, y, self.sigma, proj)

    def measure(self, data, cliques=None, weights=None, public=False, flatten=None):
        if cliques is None:
            cliques = self.cliques

        if weights is None:
            weights = np.ones(len(self.cliques))

        weights = np.array(weights) / np.linalg.norm(weights)
        if self.n_jobs > 1:
            measurements = [
                meas
                for meas in Pool(self.n_jobs).starmap(
                    self._measure,
                    zip(
                        itertools.cycle([data]),
                        cliques,
                        weights,
                        itertools.cycle([flatten]),
                        itertools.cycle([public]),
                    ),
                )
            ]
        else:
            measurements = [
                self._measure(data, clique, weight, flatten, public) for clique, weight in zip(cliques, weights)
            ]
        return measurements

    def generate(self, n_records: int = None, condition_records: pd.DataFrame = None):
        if self.fit_state != "trained":
            raise ValueError(
                "Model has not been fully trained yet. please run a .fit call with `public` set to `False`"
            )

        if condition_records is not None:
            _conditions = condition_records.copy()

        if (self.compress) and condition_records is not None:
            condition_records = self.compressor.transform(condition_records)

        synth = self.model.synthetic_data(
            rows=n_records, condition_records=condition_records
        )
        if self.compress:
            synth = self.compressor.reverse(synth)

        synth_df = synth.df
        if condition_records is not None:
            synth_df = pd.concat(
                [_conditions, synth_df.drop(columns=_conditions.columns)], axis=1
            )

        return synth_df



# Adapted from: https://github.com/ryan112358/private-pgm/blob/1da21c8b38149b05f1385b8e54116568b700b4fa/src/mbi/graphical_model.py

import pickle
import itertools
import numpy as np
import pandas as pd
import networkx as nx
from functools import reduce

from dpmm.models.base.mbi import Domain, Dataset, CliqueVector, JunctionTree


class GraphicalModel:
    def __init__(self, domain, cliques, total=1.0, elimination_order=None, prng=None):
        """Constructor for a GraphicalModel

        :param domain: a Domain object
        :param total: the normalization constant for the distribution
        :param cliques: a list of cliques (not necessarilly maximal cliques)
            - each clique is a subset of attributes, represented as a tuple or list
        :param elim_order: an elimination order for the JunctionTree algorithm
            - Elimination order will impact the efficiency by not correctness.
              By default, a greedy elimination order is used
        """
        self.domain = domain
        self.total = total
        tree = JunctionTree(domain, cliques, elimination_order, prng=prng)
        self.junction_tree = tree

        self.cliques = tree.maximal_cliques()  # maximal cliques
        self.message_order = tree.mp_order()
        self.sep_axes = tree.separator_axes()
        self.neighbors = tree.neighbors()
        self.elimination_order = tree.elimination_order
        self._prng = prng

        self.size = sum(domain.size(cl) for cl in self.cliques)
        if self.size * 8 > 4 * 10**9:
            import warnings

            message = "Size of parameter vector is %.2f GB. " % (self.size * 8 / 10**9)
            message += "Consider removing some measurements or finding a better elimination order"
            warnings.warn(message)

    def set_random_state(self, prng):
        if prng is not None:
            self._prng = prng
            self.junction_tree.set_random_state(prng)

    @property
    def prng(self):
        if self._prng is None:
            return np.random.RandomState()
        return self._prng

    def save(self, path):
        with open(path, "wb") as f:
            pickle.dump(self, f)

    @classmethod
    def load(cls, path):
        with open(path, "rb") as f:
            return pickle.load(f)

    def project(self, attrs):
        """Project the distribution onto a subset of attributes.
            I.e., compute the marginal of the distribution

        :param attrs: a subset of attributes in the domain, represented as a list or tuple
        :return: a Factor object representing the marginal distribution
        """
        # use precalculated marginals if possible
        if type(attrs) is list:
            attrs = tuple(attrs)
        if hasattr(self, "marginals"):
            for cl in self.cliques:
                if set(attrs) <= set(cl):
                    return self.marginals[cl].project(attrs)

        elim = self.domain.invert(attrs)
        elim_order = greedy_order(self.domain, self.cliques + [attrs], elim)
        pots = list(self.potentials.values())
        ans = variable_elimination_logspace(pots, elim_order, self.total)
        proj = ans.project(attrs)
        return proj

    def krondot(self, matrices):
        """Compute the answer to the set of queries Q1 x Q2 X ... x Qd, where
            Qi is a query matrix on the ith attribute and "x" is the Kronecker product
        This may be more efficient than computing a supporting marginal then multiplying that by Q.
        In particular, if each Qi has only a few rows.

        :param matrices: a list of matrices for each attribute in the domain
        :return: the vector of query answers
        """
        assert all(
            M.shape[1] == n for M, n in zip(matrices, self.domain.shape)
        ), "matrices must conform to the shape of the domain"
        logZ = self.belief_propagation(self.potentials, logZ=True)
        factors = [self.potentials[cl].exp() for cl in self.cliques]
        Factor = type(factors[0])  # infer the type of the factors
        elim = self.domain.attrs
        for attr, Q in zip(elim, matrices):
            d = Domain(["%s-answer" % attr, attr], Q.shape)
            factors.append(Factor(d, Q))
        result = variable_elimination(factors, elim)
        result = result.transpose(["%s-answer" % a for a in elim])
        return result.datavector(flatten=False) * self.total / np.exp(logZ)

    def calculate_many_marginals(self, projections):
        """Calculates marginals for all the projections in the list using
            Algorithm for answering many out-of-clique queries (section 10.3 in Koller and Friedman)

        This method may be faster than calling project many times

        :param projections: a list of projections, where
            each projection is a subset of attributes (represented as a list or tuple)
        :return: a list of marginals, where each marginal is represented as a Factor
        """

        self.marginals = self.belief_propagation(self.potentials)
        sep = self.sep_axes
        neighbors = self.neighbors
        # first calculate P(Cj | Ci) for all neighbors Ci, Cj
        conditional = {}
        for Ci in neighbors:
            for Cj in neighbors[Ci]:
                Sij = sep[(Cj, Ci)]
                Z = self.marginals[Cj]
                conditional[(Cj, Ci)] = Z / Z.project(Sij)

        # now iterate through pairs of cliques in order of distance
        pred, dist = nx.floyd_warshall_predecessor_and_distance(
            self.junction_tree.tree, weight=False
        )
        results = {}
        for Ci, Cj in sorted(
            itertools.combinations(self.cliques, 2), key=lambda X: dist[X[0]][X[1]]
        ):
            Cl = pred[Ci][Cj]
            Y = conditional[(Cj, Cl)]
            if Cl == Ci:
                X = self.marginals[Ci]
                results[(Ci, Cj)] = results[(Cj, Ci)] = X * Y
            else:
                X = results[(Ci, Cl)]
                S = set(Cl) - set(Ci) - set(Cj)
                results[(Ci, Cj)] = results[(Cj, Ci)] = (X * Y).sum(S)

        results = {
            self.domain.canonical(key[0] + key[1]): results[key] for key in results
        }

        answers = {}
        for proj in projections:
            for attr in results:
                if set(proj) <= set(attr):
                    answers[proj] = results[attr].project(proj)
                    break
            if proj not in answers:
                # just use variable elimination
                answers[proj] = self.project(proj)

        return answers

    def datavector(self, flatten=True):
        """Materialize the explicit representation of the distribution as a data vector."""
        logp = sum(self.potentials[cl] for cl in self.cliques)
        ans = np.exp(logp - logp.logsumexp())
        wgt = ans.domain.size() / self.domain.size()
        return ans.expand(self.domain).datavector(flatten) * wgt * self.total

    def belief_propagation(self, potentials, logZ=False):
        """Compute the marginals of the graphical model with given parameters

        Note this is an efficient, numerically stable implementation of belief propagation

        :param potentials: the (log-space) parameters of the graphical model
        :param logZ: flag to return logZ instead of marginals
        :return marginals: the marginals of the graphical model
        """
        beliefs = {cl: potentials[cl].copy() for cl in potentials}
        messages = {}
        for i, j in self.message_order:
            sep = beliefs[i].domain.invert(self.sep_axes[(i, j)])
            if (j, i) in messages:
                tau = beliefs[i] - messages[(j, i)]
            else:
                tau = beliefs[i]
            messages[(i, j)] = tau.logsumexp(sep)
            beliefs[j] += messages[(i, j)]

        cl = self.cliques[0]
        if logZ:
            return beliefs[cl].logsumexp()

        logZ = beliefs[cl].logsumexp()
        for cl in self.cliques:
            beliefs[cl] += np.log(self.total) - logZ
            beliefs[cl] = beliefs[cl].exp(out=beliefs[cl])

        return CliqueVector(beliefs)

    def mle(self, marginals):
        """Compute the model parameters from the given marginals

        :param marginals: target marginals of the distribution
        :param: the potentials of the graphical model with the given marginals
        """
        potentials = {}
        variables = set()
        for cl in self.cliques:
            new = tuple(variables & set(cl))
            # factor = marginals[cl] / marginals[cl].project(new)
            variables.update(cl)
            potentials[cl] = marginals[cl].log() - marginals[cl].project(new).log()
        return CliqueVector(potentials)

    def fit(self, data):
        from mbi import Factor

        assert data.domain.contains(
            self.domain
        ), "model domain not compatible with data domain"
        marginals = {}
        for cl in self.cliques:
            x = data.project(cl).datavector()
            dom = self.domain.project(cl)
            marginals[cl] = Factor(dom, x)
        self.potentials = self.mle(marginals)

    def synthetic_data(
        self, rows=None, condition_records: pd.DataFrame = None, method="round"
    ):
        """Generate synthetic tabular data from the distribution.
        Valid options for method are 'round' and 'sample'."""
        if condition_records is not None:
            total = condition_records.shape[0]
        elif rows is not None:
            total = rows
        else:
            total = int(self.total)

        cols = self.domain.attrs
        data = np.zeros((total, len(cols)), dtype=int)

        if condition_records is not None:
            df = condition_records.reindex(columns=self.domain.attrs, fill_value=0)
        else:
            df = pd.DataFrame(data, columns=cols)

        cliques = [set(cl) for cl in self.cliques]

        def synthetic_col(counts, total):
            if method == "sample":
                probas = counts / counts.sum()
                return self.prng.choice(counts.size, total, True, probas)
            counts *= total / counts.sum()
            frac, integ = np.modf(counts)
            integ = integ.astype(int)
            extra = total - integ.sum()
            if extra > 0:
                idx = self.prng.choice(counts.size, extra, False, frac / frac.sum())
                integ[idx] += 1
            vals = np.repeat(np.arange(counts.size), integ)
            self.prng.shuffle(vals)
            return vals

        order = self.elimination_order[::-1]
        if condition_records is None:
            col = order.pop(0)
            marg = self.project([col]).datavector(flatten=False)
            df.loc[:, col] = synthetic_col(marg, total)
            used = {col}
        else:
            used = set(condition_records.columns)
            order = [col for col in order if col not in used]

        for col in order:
            relevant = [cl for cl in cliques if col in cl]
            relevant = used.intersection(set.union(*relevant))
            proj = tuple(relevant)
            used.add(col)
            marg = self.project(proj + (col,)).datavector(flatten=False)

            def foo(group):
                idx = group.name
                vals = synthetic_col(marg[idx], group.shape[0])
                group[col] = vals
                return group

            if len(proj) >= 1:
                df = df.groupby(list(proj), group_keys=False).apply(foo)
            else:
                df[col] = synthetic_col(marg, df.shape[0])

        return Dataset(df, self.domain)


def variable_elimination_logspace(potentials, elim, total):
    """run variable elimination on a list of **logspace** factors"""
    k = len(potentials)
    psi = dict(zip(range(k), potentials))
    for z in elim:
        psi2 = [psi.pop(i) for i in list(psi.keys()) if z in psi[i].domain]
        phi = reduce(lambda x, y: x + y, psi2, 0)
        tau = phi.logsumexp([z])
        psi[k] = tau
        k += 1

    ans = reduce(lambda x, y: x + y, psi.values(), 0)
    return (ans - ans.logsumexp() + np.log(total)).exp()


def variable_elimination(factors, elim):
    """run variable elimination on a list of (non-logspace) factors"""
    k = len(factors)
    psi = dict(zip(range(k), factors))
    for z in elim:
        psi2 = [psi.pop(i) for i in list(psi.keys()) if z in psi[i].domain]
        phi = reduce(lambda x, y: x * y, psi2, 1)
        tau = phi.sum([z])
        psi[k] = tau
        k += 1
    return reduce(lambda x, y: x * y, psi.values(), 1)


def greedy_order(domain, cliques, elim):
    order = []
    unmarked = set(elim)
    cliques = set(cliques)
    total_cost = 0
    for k in range(len(elim)):
        cost = {}
        for a in unmarked:
            # all cliques that have a
            neighbors = list(filter(lambda cl: a in cl, cliques))
            # variables in this "super-clique"
            variables = tuple(set.union(set(), *map(set, neighbors)))
            # domain for the resulting factor
            newdom = domain.project(variables)
            # cost of removing a
            cost[a] = newdom.size()

        # find the best variable to eliminate
        a = min(cost, key=lambda a: cost[a])

        # do some cleanup
        order.append(a)
        unmarked.remove(a)
        neighbors = list(filter(lambda cl: a in cl, cliques))
        variables = tuple(set.union(set(), *map(set, neighbors)) - {a})
        cliques -= set(neighbors)
        cliques.add(variables)
        total_cost += cost[a]

    return order



from logging import getLogger
from pathlib import Path
from typing import Dict, List, Optional, Self, Tuple

import joblib
import numpy as np
import pandas as pd
from numpy.random import RandomState
from sklearn.preprocessing import OrdinalEncoder

from dpmm.processing import BINNER_DICT
from dpmm.utils import to_path

logger = getLogger("dpmm")


class TableBinner:
    def __init__(
        self,
        binner_type: str = "priv-tree",
        binner_settings: Dict = {"n_bins": "auto"},
        domain: Optional[Dict] = None,
        structural_zeros: Optional[Dict[str, Tuple]] = None,
        random_state: Optional[RandomState] = None,
    ):
        """
        Initialize the TableBinner. This class is used to bin numerical and categorical columns in a DataFrame.
        It supports two binning strategies and can handle missing values and structural zeros.
        Categorical columns are encoded using OrdinalEncoder.
        Static columns are dropped from the DataFrame.

        :param binner_type: The type of binner to use (e.g., "priv-tree").
        :param binner_settings: Settings for the binner.
        :param domain: Optional domain information for the data.
        :param structural_zeros: Optional structural zeros for specific columns.
        :param random_state: Optional random state for reproducibility.

        .. code-block:: python

            binner = TableBinner(binner_type="uniform", binner_settings={"n_bins": 10})
        """
        self.binner_type = binner_type
        self.binner_settings = binner_settings

        if domain is None:
            domain = {}
        self.domain = domain
        if random_state is None:
            random_state = RandomState()
        self.random_state = random_state

        self.binners = None
        self.dtypes = None
        self.oh_encoder = None
        self.bin_domain = None
        self.nan_columns = None
        self.static_columns = None
        self.col_order = None
        self.categories = None
        self.is_fit = False

        if structural_zeros is None:
            structural_zeros = {}
        self.set_structural_zeros(structural_zeros)

    def set_params(self, **kwargs) -> None:
        """
        Update binner settings with provided parameters.

        :param kwargs: Key-value pairs of parameters to update.

        .. code-block:: python

            binner.set_params(n_bins=15, epsilon=0.1)
        """
        self.binner_settings.update(kwargs)

    def set_domain(self, domain: Dict) -> None:
        """
        Set the domain for the binner.

        :param domain: A dictionary specifying the domain for each column.

        .. code-block:: python

            binner.set_domain({"col1": {"lower": 0, "upper": 10}})
        """
        if domain is not None:
            self.domain = domain

    def set_structural_zeros(self, structural_zeros: Dict[str, Tuple]) -> None:
        """
        Set the structural zeros for the binner.

        :param structural_zeros: A dictionary mapping columns to structural zero intervals.

        .. code-block:: python

            binner.set_structural_zeros({"col1": [(0.1, 0.2), (0.5, 0.6)]})
        """
        self.structural_zeros = structural_zeros
        if self.binners is not None:
            for col, binner in self.binners.items():
                if col in structural_zeros:
                    binner.set_structural_zeros(structural_zeros[col])

    def set_random_state(self, rnd: RandomState) -> None:
        """
        Set the random state for reproducibility.

        :param rnd: A numpy RandomState instance.

        .. code-block:: python

            rnd = RandomState(42)
            binner.set_random_state(rnd)
        """
        self.random_state = rnd

        if self.binners is not None:
            for _, binner in self.binners.items():
                binner.set_random_state(rnd)

    def get_categories(
        self, col: str, series: pd.Series, public: bool = False
    ) -> List[str]:
        """
        Get categories for a column, handling missing domain information.

        :param col: The column name.
        :param series: The pandas Series for the column.
        :param public: Whether the data is public (no privacy concerns).
        :return: A list of categories.

        .. code-block:: python

            categories = binner.get_categories("col1", df["col1"])
        """
        col_domain = self.domain.get(col, {})
        if "categories" in col_domain:
            categories = col_domain["categories"]
        else:
            if not (public):
                logger.warning(
                    f"PrivacyLeakage: No categorical domain provided for Column {col} - will be imputed."
                )
            categories = series.unique().tolist()

        categories = [cat for cat in categories if pd.notnull(cat)]

        all_numerical = all(
            [
                (
                    np.issubdtype(np.dtype(type(cat)), np.floating)
                    or np.issubdtype(np.dtype(type(cat)), np.integer)
                    or np.issubdtype(np.dtype(type(cat)), np.bool_)
                )
                for cat in categories
            ]
        )

        if all_numerical:
            categories = sorted(categories)
        else:
            categories = sorted(categories, key=lambda x: str(x))

        return categories

    def insert_col(
        self, df: pd.DataFrame, col: str, series: pd.Series
    ) -> Tuple[pd.DataFrame, str]:
        """
        Insert a new column into the DataFrame, ensuring no name conflicts.

        :param df: The input DataFrame.
        :param col: The name of the column to insert.
        :param series: The pandas Series to insert as a column.
        :return: A tuple containing the updated DataFrame and the final column name.

        .. code-block:: python

            df, col_name = binner.insert_col(df, "new_col", pd.Series([1, 2, 3]))
        """
        idx = 0
        while col in df.columns:
            col = f"{col}_{idx}"
            idx += 1

        return pd.concat([df, series.rename(col)], axis=1), col

    def fit(self, df: pd.DataFrame, public: bool = False) -> None:
        """
        Fit the binner to the DataFrame.

        :param df: The input DataFrame.
        :param public: Whether the data is public (no privacy concerns).

        .. code-block:: python

            binner.fit(df, public=True)
        """
        self.dtypes = df.dtypes
        self.binners = {}
        self.cat_encoders = {}
        self.bin_domain = {}
        self.nan_columns = {}
        self.static_columns = {}
        self.categories = {}
        self.has_nan = {}
        self.col_order = df.columns.tolist()

        # NaN Management
        for col, series in df.items():
            na_flag = series.isna()
            not_na_flag = ~na_flag
            if na_flag.any() and not_na_flag.any():
                fill_value = (
                    df.loc[not_na_flag, col]
                    .sample(n=1, random_state=self.random_state)
                    .iloc[0]
                )
                df, nan_col = self.insert_col(
                    df=df, col=f"{col}_NaN", series=na_flag.astype("category")
                )

                self.domain[nan_col] = {"categories": [False, True]}
                df.loc[na_flag, col] = fill_value
                self.nan_columns[col] = {"name": nan_col, "fill_value": fill_value}

        # Static Columns
        self.static_columns = {
            col: series.iloc[0]
            for col, series in df.items()
            if (series.nunique(dropna=False) == 1)
        }
        # Drop Static Columns
        df = df.drop(columns=[col for col in self.static_columns if col in df.columns])

        # Numerical Columns
        self.num_cols = [
            col
            for col, series in df.items()
            if ((series.dtype.kind) in "Mmfui" and (series.dropna().nunique() > 1))
        ]

        # Compute Epsilon
        if public:
            epsilon = None
        else:
            epsilon = self.binner_settings.get("epsilon", None)
            if epsilon is not None and len(self.num_cols) > 0:
                # Split the epsilon
                epsilon /= len(self.num_cols)

        # Categorical Columns
        self.cat_cols = [col for col in df.columns if col not in self.num_cols]
        for col in self.cat_cols:
            categories = self.get_categories(col, df[col], public=public)
            self.bin_domain[col] = len(categories)
            self.categories[col] = categories
            self.cat_encoders[col] = OrdinalEncoder(categories=[categories])
            self.cat_encoders[col] = OrdinalEncoder(categories=[categories])
            self.cat_encoders[col].fit(df[col].to_frame())

        for col in self.num_cols:
            bin_settings = dict(self.binner_settings)
            bin_settings["epsilon"] = epsilon
            if self.domain is not None:
                bin_settings.update(self.domain.get(col, {}))

            self.binners[col] = BINNER_DICT[self.binner_type](
                **bin_settings, rnd=self.random_state
            )
            # Set Structural Zeros
            if self.structural_zeros is not None:
                if col in self.structural_zeros:
                    self.binners[col].set_structural_zeros(self.structural_zeros[col])

            self.binners[col].fit(df[col].to_numpy())
            self.bin_domain[col] = self.binners[col].bin_domain

        self.is_fit = True

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Transform the DataFrame using the fitted binners and encoders.

        :param df: The input DataFrame.
        :return: The transformed DataFrame.

        .. code-block:: python

            transformed_df = binner.transform(df)
        """
        # Add NaN columns
        df = df.copy()
        for col, nan_col in self.nan_columns.items():
            na_flag = df[col].isna().rename(nan_col["name"])
            df.loc[na_flag, col] = nan_col["fill_value"]
            df = pd.concat([df, na_flag], axis=1)

        # Drop Static Columns
        df = df.drop(columns=[col for col in self.static_columns if col in df.columns])

        # Transform the DataFrame using the fitted binners and encoders
        dfs = [
            pd.Series(
                self.cat_encoders[col].transform(df[[col]]).squeeze(),
                index=df.index,
                name=col,
                dtype=int,
            )
            for col in self.cat_cols
            if col in df.columns
        ]
        dfs += [
            pd.Series(
                self.binners[col].transform(df[col].to_numpy()),
                index=df.index,
                name=col,
                dtype=int,
            )
            for col in self.num_cols
            if col in df.columns
        ]

        t_df = pd.concat(dfs, axis=1)
        return t_df

    @property
    def n_bins(self) -> Dict[str, int]:
        """
        Get the number of bins for each column.

        :return: A dictionary mapping column names to the number of bins.

        .. code-block:: python

            bins = binner.n_bins
        """
        return {
            col: binner.n_bins
            for col, binner in self.binners.items()
            if not isinstance(binner, dict)
        }

    @property
    def spent_epsilon(self) -> Dict[str, Optional[float]]:
        """
        Get the spent privacy budget (epsilon) for each column.

        :return: A dictionary mapping column names to the spent privacy budget.

        .. code-block:: python

            spent_eps = binner.spent_epsilon
        """
        return {
            col: binner.spent_epsilon
            for col, binner in self.binners.items()
            if not isinstance(binner, dict)
        }

    @property
    def zeros(self) -> Dict[str, List[int]]:
        """
        Get the indices of structural zero bins for each column.

        :return: A dictionary mapping column names to lists of structural zero indices.

        .. code-block:: python

            zero_indices = binner.zeros
        """
        _zeros = {}
        for col, col_zeros in self.structural_zeros.items():
            if col in self.cat_encoders:
                _zeros[col] = [
                    self.cat_encoders[col].transform([[zero]])[0, 0]
                    for zero in col_zeros
                ]
            elif col in self.binners:
                self.binners[col].set_structural_zeros(col_zeros)
                _zeros[col] = self.binners[col].zeros
        return _zeros

    def fit_transform(self, df: pd.DataFrame, public: bool = False) -> pd.DataFrame:
        """
        Fit the binner to the DataFrame and transform it.

        :param df: The input DataFrame.
        :param public: Whether the data is public (no privacy concerns).
        :return: The transformed DataFrame.

        .. code-block:: python

            transformed_df = binner.fit_transform(df, public=True)
        """
        self.fit(df, public=public)
        t_df = self.transform(df)
        return t_df

    def inverse_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Inverse transform the DataFrame to its original form.

        :param df: The transformed DataFrame.
        :return: The original DataFrame.

        .. code-block:: python

            original_df = binner.inverse_transform(transformed_df)
        """
        dfs = [
            pd.Series(
                self.cat_encoders[col].inverse_transform(df[[col]]).squeeze(),
                index=df.index,
                name=col,
            )
            for col in self.cat_cols
        ]

        dfs += [
            pd.Series(
                self.binners[col].inverse_transform(df[col].to_numpy()),
                index=df.index,
                name=col,
            )
            for col in self.num_cols
        ]
        t_df = pd.concat(dfs, axis=1)

        # Add static columns
        for col, value in self.static_columns.items():
            t_df[col] = value

        # Apply NaN values
        for col, nan_col in self.nan_columns.items():
            nan_col = nan_col["name"]
            na_flag = t_df[nan_col].astype(bool)
            t_df.loc[na_flag, col] = np.nan
            t_df = t_df.drop(columns=[nan_col])

        t_df = t_df[self.col_order]

        return t_df.astype(self.dtypes)

    @to_path
    def store(self, path: Path) -> None:
        """
        Store the binner to a file.

        :param path: The file path to store the binner.

        .. code-block:: python

            binner.store("binner.pkl")
        """
        joblib.dump(self, path)

    @classmethod
    def load(cls, path: Path) -> Self:
        """
        Load the binner from a file.

        :param path: The file path to load the binner from.
        :return: The loaded TableBinner instance.

        .. code-block:: python

            binner = TableBinner.load("binner.pkl")
        """
        return joblib.load(path)



import pandas as pd
from typing import Dict
from dpmm.models.base.base import GenerativeModel


class GraphicalGenerativeModel(GenerativeModel):
    def set_domain(self, domain: Dict):
        super().set_domain(domain)
        self.generator._domain = domain

    def set_structural_zeros(self, structural_zeros):
        self.generator.set_structural_zeros(structural_zeros)

    @property
    def model_size(self):
        return self.generator.model_size

    def check_fit(self, df):
        cls_name = self.__class__.__name__
        # check No Missing Values
        not_na = [col for col, series in df.items() if series.isna().any()]
        not_na_msg = (
            f"Columns {not_na} contains null values. Not supported by {cls_name}"
        )
        assert len(not_na) == 0, not_na_msg

        # Check All integers
        not_int = [col for col, series in df.items() if (series.dtype.kind not in "ui")]
        not_int_msg = (
            f"Columns {not_int} have non-int dtypes. Not Supported by {cls_name}"
        )
        assert len(not_int) == 0, not_int_msg

        # Check all Positive
        not_positive = [col for col, series in df.items() if (series < 0).any()]
        not_positive_msg = f"Columns {not_positive} have non positive values. Not Supported by {cls_name}"
        assert len(not_positive) == 0, not_positive_msg

        if getattr(self, "domain") is not None:
            upper = pd.Series(self.domain)
            real = df.max(axis=0)

            mismatch = real > upper

            if mismatch.any():
                to_high = real.loc[mismatch]
                raise ValueError(
                    f"Columns {to_high.index.tolist()} have values higher than provided domain."
                )

    def fit(self, df, public=False, marginals_only: bool = False):
        """
        Fit the model to the data.

        :param df: The dataset.
        :type df: pd.DataFrame
        :param public: Whether the data is public. Defaults to False.
        :type public: bool, optional

        **Example**::

            >>> fit(df, public=True)
        """
        self.generator.fit(df, public=public, marginals_only=marginals_only)

    def generate(self, n_records: int = None, condition_records: pd.DataFrame = None):
        """
        Generate synthetic data using the model.

        :param n_records: Number of records to generate. Defaults to None.
        :type n_records: int, optional
        :param condition_records: Conditional records. Defaults to None.
        :type condition_records: pd.DataFrame, optional
        :return: The generated synthetic data.
        :rtype: pd.DataFrame

        **Example**::

            >>> generate(n_records=100)
        """
        return self.generator.generate(
            n_records=n_records, condition_records=condition_records
        )
